"""
Useful functions
"""
import numpy as np

# ROS Libraries
import rospy
from std_msgs.msg import Bool
from trajectory_msgs.msg import (
    JointTrajectoryPoint,
    JointTrajectory
)

class JTrajectory(object):

    def __init__(self, pos, times, robot, contour=None):
        assert (len(pos)==len(times)), "positions and times must be same len"
        self.positions = pos
        self.time = times
        self.contour = contour
        self.robot = robot
        self.hash = hash((self.contour, tuple(self.positions[0])))


    def __hash__(self):
        """
        to use JTrajectory as keys in dicts
        """
        return self.hash


    def __eq__(self, other):
        return self.hash == other.hash


def joints_at_time(traj, t):
    """
    get the joint position at a time (relative to start of traj)
    """
    if t==0:
        return traj.positions[0]

    wp_i = bisect_left(traj.time, t)

    # linear interpolate between the waypoints
    j1 = np.array(traj.positions[wp_i-1])
    j2 = np.array(traj.positions[wp_i])
    t1 = traj.time[wp_i-1]
    t2 = traj.time[wp_i]

    return (t-t1)/(t2-t1)*(j2-j1) + j1


default_joint_names = ['shoulder_pan_joint',
 'shoulder_lift_joint',
 'elbow_joint',
 'wrist_1_joint',
 'wrist_2_joint',
 'wrist_3_joint']


def JTraj2ROS(jtraj,joint_names=default_joint_names):
    """
    convert JTrajectory object to ROS JointTrajectory
    """
    j = JointTrajectory()
    j.joint_names = joint_names

    for i in range(len(jtraj.positions)):
        jp = JointTrajectoryPoint()
        jp.positions = jtraj.positions[i]
        jp.time_from_start = rospy.Duration(jtraj.time[i])
        j.points.append(jp)

    return j


def speed_multiplier(jtraj, multiplier):
    """
    used for slowing down contour execution for testing
    """
    if isinstance(jtraj, JTrajectory):
        jtraj.time = [t*multiplier for t in jtraj.time]
    elif isinstance(jtraj, JointTrajectory):
        for p in jtraj.points:
            p.time_from_start = p.time_from_start*multiplier
            p.velocities = []
            p.accelerations = []
            p.effort = []
    return jtraj
